import { PluginPass, PluginObj } from '@babel/core';
import { NodePath } from '@babel/traverse';
import * as _babel_types from '@babel/types';
import { Expression } from '@babel/types';

type Macro = (this: Context, ...args: any[]) => Record<string, any> | void;
interface Options {
    macros?: Record<string, Macro>[];
    define?: Record<string, Context>[];
    polyfill?: string | null;
}

declare class Context {
    path: NodePath;
    parent: Context | undefined;
    define: Record<string, Context>;
    macros: Record<string, Macro>;
    uid: string;
    also: Set<Context>;
    props: Map<string, any>;
    usedBy: Set<NodePath<_babel_types.Node>>;
    children: Set<Context>;
    condition?: Expression | string;
    alternate?: Context;
    static get(from: NodePath): Context | undefined;
    constructor(path: NodePath, parent?: Context, name?: string);
}

type State = PluginPass & {
    context: Context;
    opts: Options;
};
declare function getUsing(path: NodePath): Set<Context>;
declare namespace Plugin {
    export { Context, type Macro, type Options };
}
declare function Plugin(_compiler: any, options: Options): PluginObj<State>;
type ExitCallback = (path: NodePath, key: string | number | null) => void;
declare function onExit(path: NodePath, callback: ExitCallback): void;

export { Context, type Macro, type Options, Plugin, type State, getUsing, onExit };
