"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// src/index.ts
var src_exports = {};
__export(src_exports, {
  Context: () => Context,
  Plugin: () => Plugin,
  getUsing: () => getUsing,
  onExit: () => onExit
});
module.exports = __toCommonJS(src_exports);

// src/context.ts
var CONTEXT = /* @__PURE__ */ new WeakMap();
var Context = class _Context {
  constructor(path, parent, name) {
    this.define = {};
    this.macros = {};
    this.uid = "";
    this.also = /* @__PURE__ */ new Set();
    this.props = /* @__PURE__ */ new Map();
    this.usedBy = /* @__PURE__ */ new Set();
    this.children = /* @__PURE__ */ new Set();
    CONTEXT.set(path, this);
    this.path = path;
    if (!(parent instanceof _Context))
      return;
    this.parent = parent;
    this.uid = name + "_" + hash(parent.uid);
    this.define = Object.create(parent.define);
    this.macros = Object.create(parent.macros);
    do
      if (parent.condition) {
        parent.children.add(this);
      }
    while (parent = parent.parent);
  }
  static get(from) {
    return CONTEXT.get(from);
  }
};
function hash(input = String(Math.random()), length = 3) {
  let hash2 = 0;
  if (input.length === 0)
    return "";
  for (let i = 0; i < input.length; i++) {
    const char = input.charCodeAt(i);
    hash2 = (hash2 << 5) - hash2 + char;
    hash2 &= hash2;
  }
  return Math.abs(hash2).toString(36).substring(0, length);
}

// src/types/assert.ts
var assert_exports = {};
__export(assert_exports, {
  and: () => and,
  falsy: () => falsy,
  inverseExpression: () => inverseExpression,
  isBinaryAssertion: () => isBinaryAssertion,
  isFalsy: () => isFalsy,
  isParenthesized: () => isParenthesized,
  ternary: () => ternary
});
var ASSERT_OP = /* @__PURE__ */ new Set([
  "in",
  "instanceof",
  "==",
  "===",
  "!=",
  "!==",
  ">",
  ">=",
  "<",
  "<="
]);
var INVERSE_OP = /* @__PURE__ */ new Map([
  ["==", "!="],
  ["===", "!=="],
  [">", "<="],
  ["<", ">="]
]);
for (const [a, b] of INVERSE_OP)
  INVERSE_OP.set(b, a);
function isParenthesized(node) {
  const { extra } = node;
  return extra ? extra.parenthesized === true : false;
}
function isBinaryAssertion(exp) {
  if (_default.isBinaryExpression(exp) && ASSERT_OP.has(exp.operator))
    return true;
  return false;
}
function inverseExpression(exp) {
  const inverse = INVERSE_OP.get(exp.operator);
  if (inverse)
    return _default.binaryExpression(inverse, exp.left, exp.right);
  throw new Error(`Can't invert binary comparison ${exp.operator}.`);
}
function isFalsy(exp) {
  return _default.isUnaryExpression(exp, { operator: "!" });
}
function falsy(exp) {
  return isBinaryAssertion(exp) ? inverseExpression(exp) : _default.unaryExpression("!", exp, true);
}
function and(a, b) {
  return _default.logicalExpression("&&", a, b);
}
function ternary(test, consequent, alternate) {
  return _default.conditionalExpression(test, consequent, alternate);
}

// src/types/jsx.ts
var jsx_exports = {};
__export(jsx_exports, {
  jsxTag: () => jsxTag
});
function jsxIdentifier(name) {
  return typeof name == "string" ? _default.jsxIdentifier(name) : name;
}
function jsxTag(tag, props, children) {
  const type = jsxIdentifier(tag);
  const content = children.map(jsxContent);
  const contains = content.length > 0;
  const openingElement = _default.jsxOpeningElement(type, props, !contains);
  const closingElement = contains ? _default.jsxClosingElement(type) : null;
  return _default.jsxElement(openingElement, closingElement, content, !contains);
}
function jsxContent(child) {
  if (_default.isJSXElement(child))
    return child;
  if (_default.isStringLiteral(child) && !/\{/.test(child.value))
    return _default.jsxText(child.value);
  return _default.jsxExpressionContainer(child);
}

// src/types/index.ts
var _default = { ...assert_exports, ...jsx_exports };

// src/errors.ts
var Status = {
  currentFile: void 0
};
function ParseErrors(register) {
  const Errors = {};
  for (const error in register) {
    const message = [];
    for (const segment of register[error].split(/\{(?=\d+\})/)) {
      const ammend = /(\d+)\}(.*)/.exec(segment);
      if (ammend)
        message.push(parseInt(ammend[1]), ammend[2]);
      else
        message.push(segment);
    }
    Errors[error] = (node, ...args) => {
      let quote = "";
      if ("node" in node)
        node = node.node;
      for (const slice of message)
        quote += typeof slice == "string" ? slice : args[slice - 1];
      return Status.currentFile.buildCodeFrameError(node, quote);
    };
  }
  return Errors;
}
function parseError(path, err, modiferName) {
  if (!(err instanceof Error))
    return path.hub.buildError(path.node, `Modifier "${modiferName}" failed: ${err}`);
  const stack = err.stack.split("\n    at ");
  const message = err instanceof Error ? err.message : err;
  const error = path.hub.buildError(path.node, `Modifier "${modiferName}" failed: ${message}`);
  error.stack = stack.slice(0, stack.findIndex((line) => /^parse/.test(line)) + 1).join("\n    at ");
  return error;
}

// src/arguments.ts
var Oops = ParseErrors({
  UnaryUseless: "Unary operator here doesn't do anything",
  HexNoNegative: `Hexadecimal numbers are converted into colors (#FFF) so negative sign doesn't mean anything here.
Parenthesize the number if you really need "-{1}" for some reason...`,
  ArrowNotImplemented: "Arrow Function not supported yet",
  ArgumentSpread: "Cannot parse argument spreads for modifier handlers",
  UnknownArgument: "Unknown argument while parsing for modifier.",
  MustBeIdentifier: "Only Identifers allowed here! Call expression must reference another modifier.",
  ModiferCantParse: "Illegal value in modifier",
  ElseNotSupported: "An else statement in an if modifier is not yet supported"
});
function parseArgument(element, childKey) {
  const exp = element.get("expression").node;
  const args = parse.Expression(exp, childKey);
  return Array.isArray(args) ? args : [args];
}
var parse = {
  Expression(element, childKey) {
    if (childKey)
      element = element[childKey];
    if (_default.isParenthesized(element))
      return element;
    if (element.type in this)
      return this[element.type](element);
    throw Oops.UnknownArgument(element);
  },
  Identifier(node) {
    return node.name;
  },
  BooleanLiteral(bool) {
    return bool.value;
  },
  StringLiteral(e) {
    if (e.value === "")
      return '""';
    return e.value;
  },
  NullLiteral() {
    return null;
  },
  TemplateLiteral(temp) {
    if (temp.quasis.length == 1)
      return temp.quasis[0].value.raw;
    return temp;
  },
  NumericLiteral(number, negative) {
    let { extra: { rawValue, raw } } = number;
    if (_default.isParenthesized(number) || !/^0x/.test(raw)) {
      if (raw.indexOf(".") > 0)
        return negative ? "-" + raw : raw;
      return negative ? -rawValue : rawValue;
    }
    if (negative)
      throw Oops.HexNoNegative(number, rawValue);
    return HEXColor(raw);
  },
  UnaryExpression(e) {
    const { argument, operator } = e;
    if (operator == "-" && _default.isNumericLiteral(argument))
      return this.NumericLiteral(argument, true);
    if (operator == "!" && _default.isIdentifier(argument, { name: "important" }))
      return "!important";
    throw Oops.UnaryUseless(e);
  },
  BinaryExpression(binary) {
    const { left, right, operator } = binary;
    if (operator == "-" && _default.isIdentifier(left) && _default.isIdentifier(right, { start: left.end + 1 }))
      return left.name + "-" + right.name;
    return [
      operator,
      this.Expression(binary, "left"),
      this.Expression(binary, "right")
    ];
  },
  SequenceExpression(sequence) {
    return sequence.expressions.map((x) => this.Expression(x));
  },
  CallExpression(e) {
    const callee = e.callee;
    const args = [];
    if (callee.type !== "Identifier")
      throw Oops.MustBeIdentifier(callee);
    for (const item of e.arguments) {
      if (_default.isSpreadElement(item))
        throw Oops.ArgumentSpread(item);
      if (!_default.isExpression(item))
        throw Oops.UnknownArgument(item);
      args.push(this.Expression(item));
    }
    const { name } = callee;
    if (CSS_UNITS.has(name))
      return args.map((x) => String(x) + name).join(" ");
    return callee.name + `(${args.join(", ")})`;
  },
  ArrowFunctionExpression(e) {
    throw Oops.ArrowNotImplemented(e);
  }
};
var CSS_UNITS = /* @__PURE__ */ new Set([
  "ch",
  "cm",
  "em",
  "ex",
  "in",
  "mm",
  "pc",
  "pt",
  "px",
  "rem",
  "vh",
  "vmax",
  "vmin",
  "vw"
]);
function HEXColor(raw) {
  raw = raw.substring(2);
  if (raw.length == 1)
    raw = "000" + raw;
  else if (raw.length == 2)
    raw = "000000" + raw;
  if (raw.length % 4 == 0) {
    let decimal = [];
    if (raw.length == 4)
      decimal = Array.from(raw).map((x) => parseInt(x + x, 16));
    else
      for (let i = 0; i < 4; i++)
        decimal.push(
          parseInt(raw.slice(i * 2, i * 2 + 2), 16)
        );
    decimal[3] = (decimal[3] / 255).toFixed(3);
    return `rgba(${decimal.join(", ")})`;
  }
  return "#" + raw;
}

// src/names.ts
function getNames(path) {
  const names = /* @__PURE__ */ new Map();
  const opening = path.get("openingElement");
  let tag = opening.get("name");
  while (tag.isJSXMemberExpression()) {
    names.set(tag.get("property").toString(), tag);
    tag = tag.get("object");
  }
  if (tag.isJSXIdentifier())
    names.set(tag.toString(), tag);
  opening.get("attributes").forEach((attr) => {
    if (!attr.isJSXAttribute() || attr.node.value)
      return;
    let { name } = attr.node.name;
    if (typeof name !== "string")
      name = name.name;
    names.set(name, attr);
  });
  return names;
}
function getName(path) {
  let encounteredReturn;
  while (path)
    switch (path.type) {
      case "LabeledStatement":
        return path.node.label.name;
      case "VariableDeclarator": {
        const { id } = path.node;
        return _default.isIdentifier(id) ? id.name : path.parentPath.node.kind;
      }
      case "AssignmentExpression":
      case "AssignmentPattern": {
        const { left } = path.node;
        return _default.isIdentifier(left) ? left.name : "assignment";
      }
      case "FunctionDeclaration":
        return path.node.id.name;
      case "ExportDefaultDeclaration":
        return getLocalFilename(path.hub);
      case "ArrowFunctionExpression": {
        path = path.parentPath;
        continue;
      }
      case "ReturnStatement": {
        if (encounteredReturn)
          return "return";
        encounteredReturn = path;
        const ancestry = path.getAncestry();
        const within = ancestry.find((x) => x.isFunction());
        const { node } = within;
        if ("id" in node && node.id)
          return node.id.name;
        if (_default.isObjectMethod(node)) {
          path = within.getAncestry()[2];
          continue;
        }
        if (_default.isClassMethod(node)) {
          if (node.key.type !== "Identifier")
            return "ClassMethod";
          if (node.key.name != "render")
            return node.key.name;
          const owner = within.parentPath.parentPath;
          if (owner.node.id)
            return owner.node.id.name;
          path = owner.parentPath;
          continue;
        }
        path = within.parentPath;
        continue;
      }
      case "ObjectProperty": {
        const { key } = path.node;
        return _default.isIdentifier(key) ? key.name : _default.isStringLiteral(key) ? key.value : "property";
      }
      default:
        return "element";
    }
  return "element";
}
function getLocalFilename(hub) {
  try {
    const { basename, dirname, sep: separator } = require("path");
    const url = hub.file.opts.filename;
    const [base] = basename(url).split(".");
    if (base !== "index")
      return base;
    return dirname(url).split(separator).pop();
  } catch (err) {
    return "File";
  }
}

// src/label.ts
function handleLabel(path) {
  const context = getContext(path);
  const body = path.get("body");
  let { name } = path.node.label;
  if (body.isBlockStatement()) {
    context.define[name] = new Context(path, context, name);
    return;
  }
  if (!body.isExpressionStatement())
    throw parseError(body, "Not an expression", name);
  if (!context)
    throw parseError(body, "Missing context", name);
  const args = parseArgument(body);
  try {
    const queue = [{ name, args }];
    while (queue.length) {
      const { name: name2, args: args2 } = queue.pop();
      const macro = context.macros[name2];
      const apply = (args3) => {
        context.props.set(name2, args3);
      };
      if (!macro) {
        apply(args2);
        continue;
      }
      const output = macro.apply(context, args2);
      if (!output)
        continue;
      if (Array.isArray(output)) {
        apply(output);
        continue;
      }
      if (typeof output != "object")
        throw new Error("Invalid modifier output.");
      for (const key in output) {
        let args3 = output[key];
        if (args3 === void 0)
          continue;
        if (!Array.isArray(args3))
          args3 = [args3];
        if (key === name2)
          apply(args3);
        else
          queue.push({ name: key, args: args3 });
      }
    }
  } catch (err) {
    throw parseError(body, err, name);
  }
}
function getContext(path) {
  let key = path.key;
  while (path = path.parentPath) {
    const context = Context.get(path);
    if (context instanceof Context) {
      if (key === "alternate") {
        let { alternate, parent, path: path2 } = context;
        if (!alternate) {
          alternate = new Context(path2, parent, "else");
          context.children.add(alternate);
          context.alternate = alternate;
        }
        return alternate;
      }
      return context;
    }
    if (path.isFunction())
      return createFunctionContext(path);
    if (path.isIfStatement())
      return createIfContext(path);
    key = path.key;
  }
  throw new Error("Context not found");
}
function createFunctionContext(path) {
  const name = getName(path);
  const context = getContext(path);
  const component = new Context(path, context, name);
  const body = path.get("body");
  component.define["this"] = component;
  onExit(path, () => {
    if (body.isBlockStatement() && !body.get("body").length)
      body.pushContainer("body", _default.expressionStatement(
        _default.jsxElement(
          _default.jsxOpeningElement(
            _default.jsxIdentifier("this"),
            [],
            true
          ),
          null,
          [],
          true
        )
      ));
  });
  return component;
}
function createIfContext(path) {
  const test = path.node.test;
  const name = _default.isIdentifier(test) ? test.name : "if";
  const outer = getContext(path);
  const inner = new Context(path, outer, name);
  if (_default.isStringLiteral(test)) {
    outer.children.add(inner);
    inner.uid = outer.uid;
    inner.condition = test.value;
  } else {
    outer.also.add(inner);
    inner.condition = test;
  }
  onExit(path, (path2, key) => {
    if (key == "alternate" || inner.alternate)
      return;
    if (!path2.removed)
      path2.remove();
  });
  return inner;
}

// src/index.ts
var HANDLED = /* @__PURE__ */ new WeakMap();
var USING_KEY = Symbol("expressive context");
function getUsing(path) {
  return new Set(path.getData(USING_KEY));
}
function Plugin(_compiler, options) {
  const SCOPE = /* @__PURE__ */ new WeakMap();
  Object.assign(_default, _compiler.types);
  function JSX(path) {
    if (path.getData(USING_KEY))
      return;
    let { node, parentPath: parent } = path;
    if (parent.isExpressionStatement()) {
      const block = parent.parentPath;
      if (block.isBlockStatement() && block.get("body").length == 1 && block.parentPath.isArrowFunctionExpression())
        block.replaceWith(_default.parenthesizedExpression(node));
      else
        parent.replaceWith(_default.returnStatement(node));
      path.skip();
      return;
    }
    const context = !parent.isJSXElement() && !parent.isJSXFragment() && getContext(parent);
    const scope = new Set(context ? [context] : SCOPE.get(parent));
    const using = /* @__PURE__ */ new Set();
    SCOPE.set(path, scope);
    if (path.isJSXElement())
      getNames(path).forEach((attr, name) => {
        let used = false;
        for (let { define } of scope) {
          const apply = [];
          for (let mod; mod = define[name]; define = Object.getPrototypeOf(define)) {
            apply.push(mod, ...mod.also);
            used = true;
            if (name == "this")
              break;
          }
          apply.reverse().forEach((ctx) => {
            ctx.usedBy.add(path);
            scope.add(ctx);
            using.add(ctx);
          });
        }
        if (used && attr.isJSXAttribute())
          attr.remove();
      });
    path.setData(USING_KEY, using);
    if (context === false || context.define.this !== context || context.props.size === 0 || context.usedBy.size)
      return;
    const [inserted] = path.replaceWith(
      _default.jsxElement(
        _default.jsxOpeningElement(_default.jSXIdentifier("this"), []),
        _default.jsxClosingElement(_default.jSXIdentifier("this")),
        path.isJSXElement() ? [path.node] : path.node.children
      )
    );
    inserted.setData(USING_KEY, /* @__PURE__ */ new Set([context]));
  }
  return {
    manipulateOptions(_options, parse2) {
      parse2.plugins.push("jsx");
    },
    visitor: {
      Program(path, state) {
        const context = new Context(path);
        context.uid = hash(state.filename);
        context.define = Object.assign({}, ...options.define || []);
        context.macros = Object.assign({}, ...options.macros || []);
      },
      JSXElement: JSX,
      JSXFragment: JSX,
      BlockStatement: {
        exit
      },
      LabeledStatement: {
        enter(path) {
          const body = path.get("body");
          if (body.isFor() || body.isWhile())
            return;
          handleLabel(path);
          onExit(path, () => {
            if (!path.removed)
              path.remove();
          });
        },
        exit
      }
    }
  };
}
function onExit(path, callback) {
  HANDLED.set(path, callback);
}
function exit(path) {
  for (const p of path.getAncestry()) {
    if (p.isBlockStatement())
      continue;
    const callback = HANDLED.get(p);
    if (callback)
      callback(p, p.key);
    else
      break;
  }
}
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  Context,
  Plugin,
  getUsing,
  onExit
});
